/*****************************************************************************
  Driver for the 2x40 LCD Panels
  (c) 2000 Joe Sunday
  This software is licensed under the GPL. See the file COPYING for details.
  The additional right is granted to port/link this module to any other
  Unix operating system, regardless of license, provided the source for
  the port remains available upon request, and this copyright notice remains
  intact.
****************************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/wrapper.h>
#include <linux/parport.h>
#include <linux/delay.h>
#include "lcd.h"

/* In 2.2.3 /usr/include/linux/version.h includes 
 * a macro for this, but 2.0.35 doesn't - so I add 
 * it here if necessary. */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif


/* Conditional compilation. LINUX_VERSION_CODE is 
 * the code (as per KERNEL_VERSION) of this version. */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h> /* for put_user */
#endif

#define SUCCESS 0


/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modsetver.h>
#endif


/* Device Declarations **************************** */

/* parport handle */
struct parport *myport = NULL;
struct pardevice *mypdevice = NULL;

/* Is the device open right now? Used to prevent 
 * concurent access into the same device */
static int Device_Open = 0;
/* Minor number of currently open device */
static int Minor;


/* Utility Functions */
/* Waits if device is busy. */
void busy() {
	int wait = 0;
	parport_write_control( myport, 0x28 );
	/* Wait for 1000 usec or until the busy flag goes low. The wait is
 	   so that we don't lock on non-existent lcds or parallel ports that
	   don't tri-state right. */
	while( (wait < 100 ) && (parport_read_data( myport ) & 0x80) ) {
		udelay( 10 );
		wait++;
	}
}

void sendinstr( char instr ) {
	parport_write_control( myport, 0x0a );
	parport_write_data( myport, instr );
	parport_write_control( myport, 0x0b );	
	busy();
}

void senddata( char data ) {
	parport_write_control( myport, 0x02 );
	parport_write_data( myport, data );
	parport_write_control( myport, 0x03 );
	busy();
}

void initialize() {
	sendinstr( 0x38 ); // Sends 2 line data mode
	sendinstr( 0x01 ); // Home and clear cursor
	sendinstr( 0x02 ); // Home display
	sendinstr( 0x0f ); // Turn everything off
}

void display( char* str ) {
	int i=0;
	while( str[i] != 0 ) {
		senddata( str[i] );
		i++;
	}
}
	
/* This function is called whenever a process 
 * attempts to open the device file */
static int device_open(struct inode *inode, 
                       struct file *file)
{
  /* We don't want to talk to two processes at the 
   * same time */
  if (Device_Open)
    return -EBUSY;
  if( parport_claim( mypdevice ) ) {
	return -EBUSY;
  }
  /* If this was a process, we would have had to be 
   * more careful here.
   *
   * In the case of processes, the danger would be 
   * that one process might have check Device_Open 
   * and then be replaced by the schedualer by another 
   * process which runs this function. Then, when the 
   * first process was back on the CPU, it would assume 
   * the device is still not open. 
   *
   * However, Linux guarantees that a process won't be 
   * replaced while it is running in kernel context. 
   *
   * In the case of SMP, one CPU might increment 
   * Device_Open while another CPU is here, right after 
   * the check. However, in version 2.0 of the 
   * kernel this is not a problem because there's a lock 
   * to guarantee only one CPU will be kernel module at 
   * the same time. This is bad in  terms of 
   * performance, so version 2.2 changed it. 
   * Unfortunately, I don't have access to an SMP box 
   * to check how it works with SMP.
   */

  Device_Open++;
  Minor = (inode->i_rdev & 0xFF); 
  if( Minor != 0 ) /* only one we support right now */
	return -EINVAL;

  /* Make sure that the module isn't removed while 
   * the file is open by incrementing the usage count 
   * (the number of opened references to the module, if 
   * it's not zero rmmod will fail)
   */
  MOD_INC_USE_COUNT;

  /* Initialize the lcd panel */
  initialize();

  return SUCCESS;
}


/* This function is called when a process closes the 
 * device file. It doesn't have a return value in 
 * version 2.0.x because it can't fail (you must ALWAYS
 * be able to close a device). In version 2.2.x it is 
 * allowed to fail - but we won't let it. 
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode *inode, 
                          struct file *file)
#else 
static void device_release(struct inode *inode, 
                           struct file *file)
#endif
{
  parport_release( mypdevice );

  /* We're now ready for our next caller */
  Device_Open --;

  /* Decrement the usage count, otherwise once you 
   * opened the file you'll never get rid of the module.
   */
  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}


/* This function is called whenever a process which 
 * have already opened the device file attempts to 
 * read from it. */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_read(struct file *file,
    char *buffer,    /* The buffer to fill with data */
    size_t length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
#else
static int device_read(struct inode *inode,
                       struct file *file,
    char *buffer,   /* The buffer to fill with 
                     * the data */ 
    int length)     /* The length of the buffer 
                     * (mustn't write beyond that!) */
#endif
{
  return -EINVAL;
}




/* This function is called when somebody tries to write 
 * into our device file */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t device_write(struct file *file,
    const char *buffer,    /* The buffer */
    size_t length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
#else
static int device_write(struct inode *inode,
                        struct file *file,
                        const char *buffer,
                        int length)
#endif
{
	int len = length;
	int count;
	char buf[100];
	for( count = 0; count < len && count < 100; count++ ) {
		 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
                 get_user( buf[count], buffer+count );
                #else
                 buf[count] = get_user( buffer + count );
                #endif
	}
	buf[count] = 0;
	display( buf );
	return count;
}

/* ioctl */
int device_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
	char scratch = (char)ioctl_param;
	switch( ioctl_num ) {
		case LCD_RESET:
			initialize();
			return 0;
		case LCD_CLEAR_DISPLAY:
			sendinstr( 0x01 );
			return 0;	
		case LCD_CURSOR_HOME:
			sendinstr( 0x02 );
			return 0;
		case LCD_ENTRY_MODE:
			scratch &= 0x03; // Strip off the upper bits just in case.
			scratch |= 0x04; // Set the instruction bit
			sendinstr( scratch );
			return 0;
		case LCD_DISPLAY_MODE:
			scratch &= 0x07;
			scratch |= 0x08;
			sendinstr( scratch );
			return 0;
		case LCD_DISPLAY_SHIFT:
			scratch &= 0x0C;
			scratch |= 0x10;
			sendinstr( scratch );
			return 0;
		case LCD_FUNCTION_SET:
			scratch &= 0x1C;
			scratch |= 0x20;
			sendinstr( scratch );
			return 0;
		case LCD_SET_DDRAM_ADDR:
			scratch &= 0x7F;
			scratch |= 0x80;
			sendinstr( scratch );
			return 0;
		default:
			return -EINVAL;
	}
}
/* Module Declarations ***************************** */
static int Major;

struct file_operations Fops = {
  NULL,   /* seek */
  device_read, 
  device_write,
  NULL,   /* readdir */
  NULL,   /* select */
  device_ioctl,
  NULL,   /* mmap */
  device_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,   /* flush */
#endif
  device_release  /* a.k.a. close */
};


/* Initialize the module - Register the character device */
int init_module()
{
  /* Register the character device (atleast try) */
  Major = module_register_chrdev(DEVICE_MAJOR, 
                                 DEVICE_NAME,
                                 &Fops);

  /* Negative values signify an error */
  if (Major < 0) {
    printk ("%s device failed with %d\n",
            "Sorry, registering the character",
            Major);
    return Major;
  }

  myport = parport_enumerate();
  if( myport == NULL ) {
	printk( "lcd couldn't get a parallel port\n" );
	return -1;
  }
 
  mypdevice = parport_register_device( myport, DEVICE_NAME, NULL, NULL, NULL, 0,
	NULL );
  if( mypdevice == NULL ) {
	printk( "lcd couldn't grab a pardevice\n" );
	return -1;
  }
  printk ("LCD registerd on major device number %d, parallel port at 0x%x\n",
          DEVICE_MAJOR, (unsigned int)myport->base);
 
  return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void cleanup_module()
{
  int ret;
  
  if( mypdevice ) {
	parport_unregister_device( mypdevice );
  }

  /* Unregister the device */
  ret = module_unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
 
  /* If there's an error, report it */ 
  if (ret < 0)
    printk("Error in unregister_chrdev: %d\n", ret);
}
